@@ -1,6 +1,6 @@
 #!/usr/bin/perl
 #
-#  	$Id: adduser.pl,v 1.94 1995/05/01 02:50:27 tedhajek Exp $	
+#  	$Id: adduser.pl,v 1.99 1996/05/29 15:03:06 sjp Exp $	
 #
 # adduser: a utility to add users to the system
 # addgroup: a utility to add groups to the system
@@ -65,17 +65,6 @@ $verbose = 1;
 # should we allow bad names?
 $allow_badname = 0;
 
-##
-## we must be root.
-##
-if ($> != 0) {
-  print STDERR "$progname: only root may add a user or group to the system.\n";
-  &clean_up();
-  exit 1;
-}
-
-
-
 
 ##
 ## global variables to store the states of the password and group
@@ -114,6 +103,9 @@ $default{"home"} = "/home";
 $default{"skel"} = "/etc/skel";
 $default{"usergroups"} = "yes";
 $default{"users_gid"} = "100";
+$default{"grouphomes"} = "no";
+$default{"letterhomes"} = "no";
+$default{"keep_nis_at_end"} = "yes";
 
 
 ##
@@ -180,13 +172,47 @@ while ($arg = shift(@ARGV)) {
     } elsif ($arg eq "--system") {
 	$action = "addsysuser";
     } elsif ($arg eq "--group") {
+	if ($ingroup_name) {
+	    print STDERR "adduser: --group option conflicts with --ingroup option\n\n";
+	    &usage();
+	    exit 1;
+	}
 	$found_group_opt = 1;
+    } elsif ($arg eq "--ingroup") {
+	if ($found_group_opt) {
+	    print STDERR "adduser: --ingroup option conflicts with --group option\n\n";
+	    &usage();
+	    exit 1;
+	}
+	    
+	$ingroup_name = shift(@ARGV);
+	if (! $ingroup_name) {
+	    &usage();
+	    exit 1;
+	}
+	unless (@tmp = getgrnam($ingroup_name)) {
+	    print STDERR "adduser: The group \"$ingroup_name\",specified by the --ingroup option, does\n";
+	    print STDERR "         not exist.  Please create and try again\n\n";
+	    &usage();
+	    exit 1;
+	}
+	$ingroup_gid = $tmp[2];
     } elsif ($arg eq "--home") {
 	$special_home = shift(@ARGV);
 	if (! $special_home) {
 	    &usage();
 	    exit 1;
 	}
+	if ( -d $special_home) {
+	    print STDERR "$progname: The home directory you specified already exists.\n";
+	    &clean_up();
+	    exit 1;
+	}
+	if ($special_home !~ /^\// ) {
+	    print STDERR "$progname: The home specified must be a complete path and must begin\n\twith a slash (/).\n";
+	    &clean_up();
+	    exit 1;
+	}
     } elsif ($arg =~ /^--/) {	# bad argument!
 	&usage();
 	exit 1;
@@ -195,6 +221,15 @@ while ($arg = shift(@ARGV)) {
     }
 }
 
+##
+## we must be root.
+##
+if ($> != 0) {
+  print STDERR "$progname: only root may add a user or group to the system.\n\"$progname --help\" gives more helpful information\n";
+  &clean_up();
+  exit 1;
+}
+
 if (! $names[0]) {		# hey, we need a username!
     &usage();
     exit 1;
@@ -227,18 +262,31 @@ if ($found_group_opt) {
     }
 }
 
+if (length($new_name) > 8) {
+    if ($action eq "addgroup") {
+	print "The group name must be less than 9 characters\n";
+    } else {
+	print "The user name must be less than 9 characters\n";
+    }
+    exit 1;
+}
+
+
 #####
 # OK, we've processed the arguments.  $action equals one of the following,
 # and the appropriate variables have been set:
 #
 # $action = "adduser"
 #    $new_name            - the name of the new user
+#    $special_home        - the home directory. if it's unset, use the default.
+#    $ingroup_name        - the name of the group to add the user to
 # $action = "addgroup"
 #    $new_name            - the name of the new group
 # $action = "addsysuser"
 #    $special_home        - the home directory. if it's unset, use the default.
 #    $make_group_also     - boolean, should we also create a group?
 #    $new_name            - the name of the new user
+#    $ingroup_name        - the name of the group to add the user to
 # $action = "addusertogroup"
 #    $existing_user       - the user to be added
 #    $existing_group      - the group to add her to
@@ -272,6 +320,9 @@ if ($new_name) {
 # read the file of defaults.
 %config = &get_config($DEFAULTS, %default);
 
+# override "usergroups" if we have the "--ingroup" flag.
+$config{'usergroups'} = 'no' if ($ingroup_name ne '');
+
 if ($action eq "adduser") {	# adduser
     if (&user_exists($new_name)) {
 	print STDERR "$progname: the user you specified already exists.\n";
@@ -328,8 +379,8 @@ if ($action eq "adduser" || $action eq "addsysuser") {
 	exit 1;
     }
     # lock the password file
-    link $PASSWD, $PLOCK;
     $passwd_state = "locked";
+    link $PASSWD, $PLOCK;
 }
 ##
 ## check if the group file is locked, if necessary:
@@ -344,8 +395,8 @@ if (($action eq "adduser" && $config{"usergroups"} eq "yes") ||
 	exit 1;
     }    
     # lock the group file
-    link $GROUP, $GLOCK;
     $group_state = "locked";
+    link $GROUP, $GLOCK;
 }
 
 
@@ -409,7 +460,12 @@ elsif ($action eq "addsysuser") {
 	$new_uid = &first_avail_id($min_uid,
 				   $config{"first_uid"},
 				   @current_uids);
-	$new_gid = $nogroup_id;	# put the user in "nogroup"
+	if (defined($ingroup_gid)) {
+	    $new_gid = $ingroup_gid;
+	}
+	else {
+	    $new_gid = $nogroup_id;	# put the user in "nogroup"
+	}
 	# make sure $new_gid exists
 	$new_group_name = getgrgid($new_gid);
 	print "User's group is named $new_group_name.\n" if ($debugging);
@@ -469,7 +525,8 @@ elsif ($action eq "addsysuser") {
     if ($special_home) {
 	$home_dir = $special_home;
     } else {
-	$home_dir = $config{"home"} . "/" . $new_name;
+	$home_base = &homebase($config{"home"}, $new_name, $new_group_name);
+	$home_dir = $home_base . "/" . $new_name;
     }
 
     &add_user_to_file($new_name,
@@ -508,11 +565,11 @@ elsif ($action eq "addsysuser") {
     }
 
     if ($make_group_also) {
-	chmod (02775, $home_dir);
 	chown ($new_uid, $new_gid, $home_dir);
+	chmod (02775, $home_dir);
     } else {
-	chmod (0755, $home_dir);
 	chown ($new_uid, 0, $home_dir);
+	chmod (0755, $home_dir);
     }
 
     &clean_up();
@@ -531,7 +588,7 @@ elsif ($action eq "adduser") {
 	$new_uid = &first_avail_id($config{"first_uid"},
 				   $nobody_id,
 				   @current_uids);
-	$new_gid = $config{"users_gid"};
+	$new_gid = $ingroup_name ? $ingroup_gid : $config{"users_gid"};
 	# make sure $new_gid exists
 	$new_group_name = getgrgid($new_gid);
 	print "User's group is named $new_group_name.\n" if ($debugging);
@@ -602,7 +659,13 @@ elsif ($action eq "adduser") {
     ## add the new user to the passwd file
     ##
     print "Updating password file... " if ($verbose);
-    $home_dir = $config{"home"} . "/" . $new_name;
+    if ($special_home) {
+	$home_dir = $special_home;
+    } else {
+	$home_base = &homebase($config{"home"}, $new_name, $new_group_name);
+	$home_dir = $home_base . "/" . $new_name;
+    }
+
     &add_user_to_file($new_name, 
 		      $new_uid, 
 		      $new_gid, 
@@ -651,34 +714,39 @@ elsif ($action eq "adduser") {
 	}
 	mkdir ($home_dir, $dir_mode);
 	chown ($new_uid, $new_gid, $home_dir);
+	chmod ($dir_mode, $home_dir);
 	print "done.\n" if ($verbose);
 
 	##
         ## get files from $SKEL
 	##
 	print "Copying files from $SKEL... " if ($verbose);
-	@skelfiles = &find_files($SKEL);
+	@skelfiles = &find_files_and_dirs($SKEL);
 	foreach $skelfile (@skelfiles) {
-	    &copy_to_dir($skelfile, $home_dir);
+	    &copy_to_dir($SKEL, $skelfile, $home_dir);
 	}
 
 	##
 	## change umask lines in appropriate skel files
 	## if we're using usergroups.
 	##
+	local (@statreturn);
 	if ($config{"usergroups"} eq "yes") {
 	    foreach $file (".login", ".profile", ".bash_profile") {
 		$this_file = $home_dir . "/" . $file;
 		if (-f $this_file) {
 		    open (FILE, "$this_file") || die "open: $!";
-		    open (NEWFILE, ">$file.new") || die "open: $!";
+		    open (NEWFILE, ">$this_file.new") || die "open: $!";
 		    while ($line = <FILE>) {
 			$line =~ s/umask 0([267])\1/umask 00$1/;
-			print NEWFILE $line;
+			print (NEWFILE $line) || die "write: $!";
 		    }
+                    (@statreturn= stat(FILE)) || die "fstat: $!";
+                    $filemode= $statreturn[2];
+                    chmod($statreturn[2],"$this_file.new") || die "chmod: $!";
 		    close FILE;
-		    close NEWFILE;
-		    rename ("$file.new", "$file") || die "rename: $!";
+		    close(NEWFILE) || die "close: $!";
+		    rename ("$this_file.new", "$this_file") || die "rename: $!";
 		}
 	    }
 	}
@@ -688,7 +756,9 @@ elsif ($action eq "adduser") {
 	##
 	if ($home_dir ne "/") {	# we really don't want to do this to root
 	    @homefiles = &find_files_and_dirs($home_dir);
-	    chown ($new_uid, $new_gid, @homefiles);
+	    foreach $homefile (@homefiles) {
+		chown ($new_uid, $new_gid, "$home_dir/$homefile");
+	    }
 	}
 	print "done.\n" if ($verbose);
     }
@@ -698,27 +768,49 @@ elsif ($action eq "adduser") {
     ##
     ($the_name, $the_passwd) = getpwnam ($new_name);
     while ($the_passwd eq "*") {
-	system "passwd $new_name";
+	if ((system "passwd $new_name") != 0) {
+	    print STDERR "$progname: System call to passwd program failed\n";
+	    &clean_up;
+	    exit 1;
+	}
         ($the_name, $the_passwd) = getpwnam ($new_name);
     }
 
     ##
     ## change the finger information
     ## 
-    system "chfn $new_name";
+    if ((system "chfn $new_name") != 0) {
+	print STDERR "$progname: System call to chfn program failed\n";
+	&clean_up;
+	exit 1;
+    }
     while (1) {
 	if (-x "/usr/bin/finger") {
-	    system "/usr/bin/finger $new_name";
+	    if ((system "/usr/bin/finger $new_name") != 0) {
+		print STDERR "$progname: System call to finger program failed\n";
+		&clean_up;
+		exit 1;
+	    }
 	    print "Is this information correct [y/n]? ";
 	} else {
 	    print "Is this finger information correct [y/n]? ";
 	}
 	chop ($answer = <STDIN>);
 	last if ($answer eq "y" || $answer eq "Y");
-	system "chfn $new_name";
+	if ((system "chfn $new_name") != 0) {
+	    print STDERR "$progname: System call to chfn program failed\n";
+	    &clean_up;
+	    exit 1;
+	}
     }
     print "done.\n";
     &clean_up();
+    if (-f "/usr/local/sbin/adduser.local") {
+	exec("/usr/local/sbin/adduser.local",
+             $new_name, $new_uid, $new_gid, $home_dir);
+        die "exec adduser.local: $!";
+    }
+
     exit 0;
 }
 
@@ -730,7 +822,39 @@ elsif ($action eq "adduser") {
 ### Subroutines
 ###
 
-    
+#####################################
+## homebase
+#####################################
+sub homebase {
+    local ($letter);
+    local ($dir);
+    local ($default_dir_mode) = 0755;
+
+    # See if base directory is present.
+    return $dir if (! -d $_[0]);
+
+    # Get first letter of username.
+    $letter = substr($_[1], 0, 1);
+
+    # Find out directory name.
+    if ($config{'letterhomes'} =~ m/yes/i) {
+       $dir = $_[0] . '/' . $letter;
+    } elsif ($config{'grouphomes'} =~ m/yes/i) {
+       $dir = $_[0]. '/' . $_[2];
+    } else {
+       $dir = $_[0];
+    }
+
+    # Create directory if needed.
+    if (! -d $dir) {
+       mkdir($dir, $default_dir_mode);
+    }
+
+    # Return directory.
+    $dir;
+}
+
+
 #####################################
 ## mktree
 #####################################
@@ -856,22 +980,39 @@ sub add_user_to_group {
 ## copy_to_dir
 ##
 sub copy_to_dir {
-    local ($file, $dir) = @_;
-
-    if ((! -f $file) || (! -d $dir)) {
+    local ($fromdir, $file, $todir) = @_;
+    local (@statreturn,$filemode);
+    
+    if ( ((! -f "$fromdir/$file") && (! -d "$fromdir/$file")) || 
+	(! -d $todir)) {
 	return -1;
     }
 
-    open (FILE, $file) || die "open: $!";
-    $file =~ s+.*/++;		# get the basename
-    open (NEWFILE, ">$dir/$file") || die "open: $!";
+    if ( -f "$fromdir/$file") {
+	open (FILE, "$fromdir/$file") || die "open: $!";
+	open (NEWFILE, ">$todir/$file") || die "open: $!";
+	
+	while (<FILE>) {
+	    print(NEWFILE) || die "print: $!";
+	}
 
-    while (<FILE>) {
-	print NEWFILE;
+	close FILE;
+	close(NEWFILE)  || die "close: $! ($todir/$file)";
+
+    } else {
+	if ( ! -d "$todir/$file") {
+	    mkdir("$todir/$file", 755) || die "mkdir: $!";
+	}
     }
 
-    close FILE;
-    close NEWFILE;
+    (@statreturn= stat("$fromdir/$file")) || die "stat: $!";
+    $filemode= $statreturn[2];
+    if ($config{"usergroups"} eq "yes") {
+        $filemode= ($filemode & 0707) | (($filemode & 0700)>>3);
+    } else {
+        $filemode= ($filemode & 0707) | (($filemode & 0007)<<3);
+    }
+    chmod($filemode,"$todir/$file") || die "chmod: $!";
 
     return 1;
 }
@@ -885,21 +1026,24 @@ sub find_files {
     local (@file_list);
 
     if (! -d $dir) {
-	return @file_list;
+        return @file_list;
     }
 
     open (FIND, "/usr/bin/find $dir -type f -print |");
     while (<FIND>) {
-	chop;
-	push (@file_list, $_);
+        chop;
+        push (@file_list, $_);
     }
     close FIND;
     return @file_list;
 }
 
+
 ##
 ## find_files_and_dirs
 ## 
+# returns file names relative to $dir
+#
 sub find_files_and_dirs {
     local ($dir) = @_;
     local (@file_list);
@@ -908,7 +1052,7 @@ sub find_files_and_dirs {
 	return @file_list;
     }
 
-    open (FIND, "/usr/bin/find $dir -print |");
+    open (FIND, "(cd $dir; /usr/bin/find . -print )|");
     while (<FIND>) {
 	chop;
 	push (@file_list, $_);
@@ -944,6 +1088,10 @@ sub add_user_to_file {
 	chop $line;
 	($name, $passwd, $uid, $gid, $quota, $comment,
         	$gcos, $dir, $shell) = split(/:/, $line);
+	if ($config{'keep_nis_at_end'} ne 'no') {
+	    # Leave NIS entries at the end.
+	    $uid = $new_uid if ($name =~ m/^\+/);
+	}
 	if ($uid < $new_uid || $added_new_one || $uid == $nobody_id) {
 	    print PASSWD $line, "\n";
 	    print STDOUT "UID: $uid\n" if $debugging;
@@ -1023,7 +1171,7 @@ sub first_avail_id {
 
     # find the first available ID
     $last_one = shift(@sorted_ids);
-    while ($this_one = shift(@sorted_ids)) {
+    while (defined($this_one = shift(@sorted_ids))) {
 	if ($this_one <= $min) {
 	    $last_one = $this_one; # advance if we're under $min
 	    next;
@@ -1115,6 +1263,12 @@ sub get_config {
 	    $config{"usergroups"} = $1;
 	} elsif ($current_line =~ /^USERS_GID\s*=\s*(\S+)/) {
 	    $config{"users_gid"} = $1;
+	} elsif ($current_line =~ /^GROUPHOMES\s*=\s*(\S+)/) {
+	    $config{"grouphomes"} = $1;
+	} elsif ($current_line =~ /^LETTERHOMES\s*=\s*(\S+)/) {
+	    $config{"letterhomes"} = $1;
+	} elsif ($current_line =~ /^KEEP_NIS_AT_END\s*=\s*(\S+)/) {
+	    $config{"keep_nis_at_end"} = $1;
 	}
 
 	##
@@ -1202,7 +1356,7 @@ sub user_exists {
 ########################
 sub version {
     print STDERR "$progname: add a user or a group to the system.\n";
-    print STDERR " version 	$Id: adduser.pl,v 1.94 1995/05/01 02:50:27 tedhajek Exp $	\n";
+    print STDERR " version 	\$Id: adduser.pl,v 1.99 1996/05/29 15:03:06 sjp Exp $	\n";
     print STDERR "\n";
     print STDERR "Copyright (C) 1995 by Ted Hajek <tedhajek\@boombox.micro.umn.edu>\n";
     print STDERR "                  and Ian Murdock <imurdock\@gnu.ai.mit.edu>\n";
@@ -1226,7 +1380,7 @@ sub usage {
     print STDERR "\n";
     print STDERR "usage:\n";
     print STDERR "\n";
-    print STDERR "  adduser user\n";
+    print STDERR "  adduser [--home directory] [--ingroup group] user\n";
     print STDERR "    Add a user.\n";
     print STDERR "  adduser --group group\n";
     print STDERR "  addgroup group\n";
